from collections import defaultdict
import math
import random
import time

from loguru import logger
import pandas as pd
import redis
import rq
from rq.registry import StartedJobRegistry, FinishedJobRegistry

from .distributed_worker import ResilientWorker
from .utilities import chunks


class QueueManager:

    retries_before_fail = 10
    backoff = 30

    def __init__(self, name, connection):
        self._name = name
        self._logger = logger.bind(queue=name)

        self._queue = rq.Queue('vivarium', connection=connection)
        self._wip = StartedJobRegistry('vivarium', connection=connection, job_class=self._queue.job_class)
        self._finished = FinishedJobRegistry('vivarium', connection=connection, job_class=self._queue.job_class)

        self._status = {'total': 0, 'pending': 0, 'running': 0, 'failed': 0, 'finished': 0, 'done': 0., 'workers': 0}
        self._failed = False
        self._completed = False

        self._retries = QueueManager.retries_before_fail
        self._backoff = QueueManager.backoff

    @property
    def name(self):
        return self._name

    @property
    def failed(self):
        return self._failed

    @property
    def completed(self):
        return self._completed

    @property
    def jobs_to_finish(self):
        return not (self.failed or self.completed)

    def enqueue(self, jobs):
        self._logger.info(f'Enqueuing jobs in queue {self.name}')
        for job in jobs:
            # TODO: might be nice to have tighter ttls but it's hard to predict how long our jobs
            # will take from model to model and the entire system is short lived anyway
            self._queue.enqueue('vivarium_cluster_tools.psimulate.distributed_worker.worker',
                                parameters=job,
                                ttl=60 * 60 * 24 * 2,
                                result_ttl=60 * 60,
                                job_timeout='7d')

    def get_results(self):
        self._logger.info(f'Checking queue {self.name}')
        finished_jobs = self._get_finished_jobs()
        start = time.time()
        results = []
        for job_id in finished_jobs:
            result = self._get_result(job_id)
            if result is not None:
                results.append(result)
        self._logger.info(f'Retrieved {len(results)} results from queue {self.name} in {time.time() - start:.2f}s')
        return results

    def update_and_report(self):
        self._update_status()
        template = (f"Queue {self.name} - Total jobs: {{total}}, % Done: {{done:.2f}}% "
                    f"Pending: {{pending}}, Running: {{running}}, Failed: {{failed}}, Finished: {{finished}} "
                    f"Workers: {{workers}}.")
        if not (self.completed or self.failed):
            self._logger.info(template.format(**self._status))
        return self._status

    def _update_status(self):
        if self.jobs_to_finish and self._retries > 0:
            self._logger.info(f'Updating status for queue {self.name}')
            # TODO: Sometimes there are duplicate job_ids, why?
            try:
                q_pending = len(set(self._queue.job_ids)) if self._queue else 0
                q_running = len(self._wip) if self._wip else 0
                q_to_write = len(self._finished) if self._finished else 0
                q_failed = len(self._queue.failed_job_registry) if self._queue else self._status['failed']
                q_finished = self._status['finished']
                q_total = q_pending + q_running + q_failed + q_to_write + q_finished
                q_workers = len(ResilientWorker.all(queue=self._queue)) if self._queue else 0

                self._status['pending'] = q_pending
                self._status['running'] = q_running + q_to_write
                self._status['failed'] = q_failed
                self._status['finished'] = q_finished
                self._status['total'] = q_total
                self._status['workers'] = q_workers
                self._status['done'] = q_finished / q_total * 100

                if sum([q_pending, q_running, q_to_write]) == 0:
                    self._mark_complete()
                elif sum([q_pending, q_workers, q_to_write]) == 0:
                    self._logger.info(f'Queue {self.name} ran out of workers with running jobs.  Marking finished.')
                    self._mark_complete()
            except redis.connection.ConnectionError:
                self._sleep_on_it()
                self._update_status()
        else:
            self._mark_failed()

    def _get_finished_jobs(self):
        if self._retries:
            try:
                return self._finished.get_job_ids()
            except redis.exceptions.ConnectionError:
                self._sleep_on_it()
                return self._get_finished_jobs()
        else:
            self._mark_failed()
            return []

    def _get_result(self, job_id):
        job = self._get_job(job_id)
        result = None
        if job is not None:
            start = time.time()
            result = job.result
            end = time.time()
            self._logger.debug(f'Read {job_id} from msgpack from queue {self.name} in {end - start:.2f}s.')
            self._status['finished'] += 1
            self._remove_finished_job(job)
        return result

    def _get_job(self, job_id):
        if self._retries:
            try:
                start = time.time()
                job = self._queue.fetch_job(job_id)
                end = time.time()
                self._logger.debug(f'Fetched job {job_id} from queue {self.name} in {end - start:.2f}s')
                return job
            except redis.exceptions.ConnectionError:
                self._sleep_on_it()
                return self._get_job(job_id)
        else:
            self._mark_failed()
            return None

    def _remove_finished_job(self, job):
        if self._retries:
            try:
                self._finished.remove(job)
            except redis.exceptions.ConnectionError:
                self._sleep_on_it()
                self._remove_finished_job(job)
        else:
            self._mark_failed()

    def _sleep_on_it(self):
        backoff = self._backoff * random.random()
        self._logger.warning(f'Failed to connect to redis for queue {self.name}.  Retrying in {backoff}s.')
        self._retries -= 1
        time.sleep(backoff)

    def _mark_failed(self):
        # TODO: Probably should cleanup redis stuff, but not sure how yet.
        if not (self.failed or self.completed):
            self._logger.warning(f'Queue {self.name} has run out of retries.  Marking as failed.')
            self._status['failed'] += self._status['pending'] + self._status['running']
            self._status['pending'] = 0
            self._status['running'] = 0
            self._status['workers'] = 0
            self._failed = True

    def _mark_complete(self):
        if not (self.failed or self.completed):
            self._logger.info(f'All jobs on queue {self.name} complete.')
            self._status['finished'] += self._status['pending'] + self._status['running']
            self._status['pending'] = 0
            self._status['running'] = 0
            self._status['workers'] = 0
            self._completed = True


class RegistryManager:

    def __init__(self, redis_processes, already_completed):
        self._logger = logger.bind(queue='all')
        self._logger.info('Building registries.')
        self._queues = [QueueManager(i, connection=redis.Redis(hostname, port))
                        for i, (hostname, port) in enumerate(redis_processes)]
        self._previously_completed = already_completed

    @property
    def jobs_to_finish(self):
        return any([q.jobs_to_finish for q in self._queues])

    def enqueue(self, jobs):
        workers_per_queue = int(math.ceil(len(jobs) / len(self._queues)))
        for queue, jobs_chunk in zip(self._queues, chunks(jobs, workers_per_queue)):
            queue.enqueue(jobs_chunk)

    def get_results(self):
        to_check = [q for q in self._queues if q.jobs_to_finish]
        results = []
        for queue in to_check:
            results.extend(queue.get_results())
        return results

    def update_and_report(self):
        status = defaultdict(int)
        for queue in self._queues:
            queue_status = queue.update_and_report()
            for k, v in queue_status.items():
                status[k] += v
        status['done'] = status['finished'] / status['total'] * 100
        status['finished'] += self._previously_completed

        template = ("Queue all - Total jobs: {total}, % Done: {done:.2f}% "
                    "Pending: {pending}, Running: {running}, Failed: {failed}, Finished: {finished} "
                    "Workers: {workers}.")

        self._logger.info(template.format(**status))
        return status

    def __len__(self):
        return len(self._queues)



from pathlib import Path

import click
from loguru import logger

from vivarium.framework.utilities import handle_exceptions
from vivarium_cluster_tools.psimulate import globals as vct_globals, runner, utilities

shared_options = [
    click.option('--project', '-P',
                 type=click.Choice(vct_globals.CLUSTER_PROJECTS),
                 default=vct_globals.DEFAULT_CLUSTER_PROJECT,
                 help='The cluster project under which to run the simulation.'),
    click.option('--queue', '-q',
                 type=click.Choice(['all.q', 'long.q']),
                 default=None,  # dynamically set based on max-runtime
                 help='The cluster queue to assign psimulate jobs to. Queue defaults to the '
                      'appropriate queue based on max-runtime. long.q allows for much longer '
                      'runtimes but there may be reasons to send jobs to that queue even '
                      'if they don\'t have runtime constraints, such as node availability.'),
    click.option('--peak-memory', '-m',
                 type=int,
                 default=3,
                 help=('The estimated maximum memory usage in GB of an individual simulate job. '
                       'The simulations will be run with this as a limit.')),
    click.option('--max-runtime', '-r',
                 type=str,
                 default='24:00:00',
                 help=('The estimated maximum runtime (DD:HH:MM) of the simulation jobs. '
                       'By default, the cluster will terminate jobs after 24h regardless of '
                       'queue. The maximum supported runtime is 3 days. Keep in mind that the '
                       'session you are launching from must be able to live at least as long '
                       'as the simulation jobs, and that runtimes by node vary wildly.')),
    click.option('--pdb', 'with_debugger',
                 is_flag=True,
                 help='Drop into python debugger if an error occurs.'),
    click.option('--redis',
                 type=int,
                 default=-1,
                 help=(f'Number of redis databases to use.  Defaults to a redis instance for every '
                       f'{vct_globals.DEFAULT_JOBS_PER_REDIS_INSTANCE} jobs.')),
    click.option('-v', 'verbose',
                 count=True,
                 help='Configure logging verbosity.'),
    click.option('--no-batch',
                 is_flag=True,
                 help="Don't batch results, write them as they come in."),
    click.option('--no-cleanup',
                 is_flag=True,
                 hidden=True,
                 help="Hidden developer option, if flagged, don't automatically cleanup results directory on failure.")
]


def pass_shared_options(func):
    # add all the shared options to the command
    for option in shared_options:
        func = option(func)
    return func


@click.group()
def psimulate():
    """A command line utility for running many simulations in parallel.

    You may initiate a new run with the ``run`` sub-command or restart a run
    from where it was stopped by using the ``restart`` sub-command.

    """
    pass


@psimulate.command()
@click.argument('model_specification', type=click.Path(exists=True, dir_okay=False))
@click.argument('branch_configuration', type=click.Path(exists=True, dir_okay=False))
@click.option('--artifact_path', '-i', type=click.Path(resolve_path=True), help='The path to the artifact data file.')
@click.option('--result-directory', '-o', type=click.Path(exists=True, file_okay=False), default=None,
              help='The directory to write results to. A folder will be created in this directory with the same name '
                   'as the configuration file.')
@pass_shared_options
def run(model_specification, branch_configuration, artifact_path, result_directory, **options):
    """Run a parallel simulation.

    The simulation itself is defined by a MODEL_SPECIFICATION yaml file
    and the parameter changes across runs are defined by a BRANCH_CONFIGURATION
    yaml file.

    The path to the data artifact can be provided as an argument here, in the
    branch configuration, or in the model specification file. Values provided as
    a command line argument or in the branch specification file will override a
    value specified in the model specifications file. If an artifact path is
    provided both as a command line argument and to the branch configuration file
    a ConfigurationError will be thrown.

    If a results directory is provided, a subdirectory will be created with the
    same name as the MODEL_SPECIFICATION if one does not exist. Results will be
    written to a further subdirectory named after the start time of the
    simulation run.

    If a results directory is not provided the base results_directory is taken
    to be /share/costeffectiveness/results.

    """
    utilities.configure_master_process_logging_to_terminal(options['verbose'])
    main = handle_exceptions(runner.main, logger, options['with_debugger'])

    main(model_specification, branch_configuration, artifact_path, result_directory,
         {'project': options['project'],
          'queue': options['queue'],
          'peak_memory': options['peak_memory'],
          'max_runtime': options['max_runtime']},
         redis_processes=options['redis'], no_batch=options['no_batch'], no_cleanup=options['no_cleanup'])


@psimulate.command()
@click.argument('results-root', type=click.Path(exists=True, file_okay=False, writable=True))
@pass_shared_options
def restart(results_root, **options):
    """Restart a parallel simulation from a previous run at RESULTS_ROOT.

    Restarting will not erase existing results, but will start workers to
    perform the remaining simulations.  RESULTS_ROOT is expected to be an
    output directory from a previous ``psimulate run`` invocation.

    """
    utilities.configure_master_process_logging_to_terminal(options['verbose'])
    main = handle_exceptions(runner.main, logger, options['with_debugger'])

    main(None, None, None, results_root,
         {'project': options['project'],
          'queue': options['queue'],
          'peak_memory': options['peak_memory'],
          'max_runtime': options['max_runtime']},
         redis_processes=options['redis'], restart=True, no_batch=options['no_batch'], no_cleanup=options['no_cleanup'])


@psimulate.command()
@click.argument('results-root', type=click.Path(exists=True, file_okay=False, writable=True))
@click.option('--add-draws', type=int, default=0, help='The number of input draws to add to a previous run.')
@click.option('--add-seeds', type=int, default=0, help='The number of random seeds to add to a previous run.')
@pass_shared_options
def expand(results_root, **options):
    """Expand a previous run at RESULTS_ROOT by adding input draws and/or
    random seeds.

    Expanding will not erase existing results, but will start workers to perform
    the additional simulations determined by the added draws/seeds.
    RESULTS_ROOT is expected to be an output directory from a previous
    ``psimulate run`` invocation.

    """
    utilities.configure_master_process_logging_to_terminal(options['verbose'])
    main = handle_exceptions(runner.main, logger, options['with_debugger'])

    main(None, None, None, results_root,
         {'project': options['project'],
          'queue': options['queue'],
          'peak_memory': options['peak_memory'],
          'max_runtime': options['max_runtime']},
         redis_processes=options['redis'],
         restart=True,
         expand={'num_draws': options['add_draws'],
                 'num_seeds': options['add_seeds']},
         no_batch=options['no_batch'],
         no_cleanup=options['no_cleanup'])



import datetime
import json
import math
import os
import random
from pathlib import Path
from time import sleep, time
from traceback import format_exc
from typing import Mapping

import numpy as np
import pandas as pd
import redis
from loguru import logger
from rq import Queue, get_current_job
from rq.job import JobStatus
from rq.registry import FailedJobRegistry
from rq.worker import StopRequested, Worker

from vivarium_cluster_tools.psimulate import globals as vct_globals
from vivarium_cluster_tools.vipin.perf_counters import CounterSnapshot


def retry_handler(job, *exc_info):
    retries = job.meta.get('remaining_retries', 2)

    if retries > 0:
        retries -= 1
        job.meta['remaining_retries'] = retries
        job.set_status(JobStatus.QUEUED)
        job.exc_info = exc_info
        job.save()
        q = Queue(name=job.origin, connection=job.connection)
        q.enqueue_job(job)
        logger.info(f'Retrying job {job.id}')
    else:
        logger.error(f'Failing job {job.id}')
        # exc_string = Worker._get_safe_exception_string(traceback.format_exception(*exc_info))
        q = Queue(name=job.origin, connection=job.connection)
        failed_queue = FailedJobRegistry(queue=q)
        failed_queue.add(job, exc_string=exc_info)
    return False


class ResilientWorker(Worker):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.acceptable_failure_count = 3

    def work(self, *args, **kwargs):
        worker_ = self.name
        logger.remove()
        logging_directory = Path(os.environ['VIVARIUM_LOGGING_DIRECTORY'])
        logger.add(logging_directory / (str(worker_) + '.log'), level='DEBUG', serialize=True)

        retries = 0
        while retries < 10:
            try:
                super(ResilientWorker, self).work(*args, **kwargs)
                return
            except redis.exceptions.ConnectionError:
                backoff = random.random()*60
                logger.error(f"Couldn't connect to redis. Retrying in {backoff}...")
                retries += 1
                sleep(backoff)
        logger.error(f"Ran out of retries. Killing worker")

    def main_work_horse(self, job, queue):
        retries = 0
        while retries < 10:
            try:
                super(ResilientWorker, self).main_work_horse(job, queue)
                return
            except redis.exceptions.ConnectionError:
                backoff = random.random()*60
                logger.error(f"Couldn't connect to redis. Retrying in {backoff}...")
                retries += 1
                sleep(backoff)
        logger.error(f"Ran out of retries. Killing work horse")

    def _monitor_work_horse_tick(self, job):
        _, ret_val = os.waitpid(self._horse_pid, 0)
        if ret_val == os.EX_OK:  # The process exited normally.
            return
        job_status = job.get_status()
        if job_status is None:  # Job completed and its ttl has expired
            return
        if job_status not in [JobStatus.FINISHED, JobStatus.FAILED]:

            if not job.ended_at:
                job.ended_at = datetime.datetime.utcnow()

            self.handle_job_failure(job=job)

            # Unhandled failure
            logger.error(f'(work-horse terminated unexpectedly; waitpid returned {ret_val})')

            retry_handler(job, f"Work-horse process was terminated unexpectedly (waitpid returned {ret_val})",
                          None, None)

            self.acceptable_failure_count -= 1
            if self.acceptable_failure_count < 0:
                logger.error("This worker has had multiple jobs fail. That may mean the host is sick. Terminating "
                             "the worker to try and shift load onto other hosts.")
                raise StopRequested()


def worker(parameters: Mapping):
    node = f"{os.environ['SGE_CLUSTER_NAME']}:{os.environ['HOSTNAME']}"
    job = f"{os.environ['JOB_NAME']}: {os.environ['JOB_ID']}:{os.environ['SGE_TASK_ID']}"

    input_draw = parameters['input_draw']
    random_seed = parameters['random_seed']
    model_specification_file = parameters['model_specification_file']
    branch_config = parameters['branch_configuration']
    logger.info(f'Launching new job {job} on {node}')
    logger.info('Starting job: {}'.format((input_draw, random_seed, model_specification_file, branch_config)))

    try:
        np.random.seed([input_draw, random_seed])
        worker_ = get_current_job().id

        from vivarium.framework.engine import SimulationContext
        from vivarium.framework.utilities import collapse_nested_dict

        configuration = {}
        run_key = {'input_draw': input_draw, 'random_seed': random_seed}

        if branch_config is not None:
            configuration.update(dict(branch_config))
            run_key.update(dict(branch_config))

        input_data_config = {
            'input_draw_number': input_draw,
        }
        if vct_globals.ARTIFACT_PATH_KEY in dict(branch_config).get(vct_globals.INPUT_DATA_KEY, {}):
            input_data_config.update({
                vct_globals.ARTIFACT_PATH_KEY: branch_config[vct_globals.INPUT_DATA_KEY][vct_globals.ARTIFACT_PATH_KEY]
            })

        configuration.update({
            'run_configuration': {
                'input_draw_number': input_draw,
                'run_id': str(worker_) + '_' + str(time()),
                'results_directory': parameters['results_path'],
                'run_key': run_key,
            },
            'randomness': {
                'random_seed': random_seed,
                'additional_seed': input_draw,
            },
            vct_globals.INPUT_DATA_KEY: input_data_config
        })

        sim = SimulationContext(model_specification_file, configuration=configuration)
        logger.info('Simulation configuration:')
        logger.info(str(sim.configuration))

        start_time = pd.Timestamp(**sim.configuration.time.start.to_dict())
        end_time = pd.Timestamp(**sim.configuration.time.end.to_dict())
        step_size = pd.Timedelta(days=sim.configuration.time.step_size)
        num_steps = int(math.ceil((end_time - start_time)/step_size))

        start_snapshot = CounterSnapshot()
        event = {'start': time()}  # timestamps of application events
        logger.info('Beginning simulation setup.')
        sim.setup()
        event['simulant_initialization_start'] = time()
        exec_time = {'setup_minutes': (event["simulant_initialization_start"] - event["start"]) / 60}  # execution event
        logger.info(f'Simulation setup completed in {exec_time["setup_minutes"]:.3f} minutes.')

        sim.initialize_simulants()
        event['simulation_start'] = time()
        exec_time['simulant_initialization_minutes'] = (
                (event["simulation_start"] - event["simulant_initialization_start"]) / 60)
        logger.info(f'Simulant initialization completed in {exec_time["simulant_initialization_minutes"]:.3f} minutes.')

        logger.info(f'Starting main simulation loop with {num_steps} time steps')
        sim.run()
        event['results_start'] = time()
        exec_time['main_loop_minutes'] = (
                (event["results_start"] - event["simulation_start"]) / 60)
        exec_time['step_mean_seconds'] = (
                (event["results_start"] - event["simulation_start"]) / num_steps)
        logger.info(f'Simulation main loop completed in {exec_time["main_loop_minutes"]:.3f} minutes.')
        logger.info(f'Average step length was {exec_time["step_mean_seconds"]:.3f} seconds.')

        sim.finalize()
        metrics = sim.report()
        event['end'] = time()
        end_snapshot = CounterSnapshot()

        do_sim_epilogue(start_snapshot, end_snapshot, event, exec_time, parameters)

        idx = pd.MultiIndex.from_tuples([(input_draw, random_seed)], names=['input_draw_number', 'random_seed'])
        output_metrics = pd.DataFrame(metrics, index=idx)
        for k, v in collapse_nested_dict(run_key):
            output_metrics[k] = v
        return output_metrics

    except Exception:
        logger.exception('Unhandled exception in worker')
        job = get_current_job()
        job.meta['root_exception'] = format_exc()
        job.save_meta()
        raise
    finally:
        logger.info('Exiting job: {}'.format((input_draw, random_seed, model_specification_file, branch_config)))


def do_sim_epilogue(start: CounterSnapshot, end: CounterSnapshot, event: dict, exec_time: dict, parameters: Mapping):
    exec_time['results_minutes'] = (event['end'] - event["results_start"]) / 60
    logger.info(f'Results reporting completed in {exec_time["results_minutes"]:.3f} minutes.')
    exec_time['total_minutes'] = (event['end'] - event["start"]) / 60
    logger.info(f'Total simulation run time {exec_time["total_minutes"]:.3f} minutes.')

    perf_log = logger.add(
        Path(os.environ['VIVARIUM_LOGGING_DIRECTORY']) / f'perf.{os.environ["JOB_ID"]}.{os.environ["SGE_TASK_ID"]}.log',
        level='DEBUG', serialize=True)
    logger.debug(json.dumps({
        "host": os.environ['HOSTNAME'].split('.')[0],
        "job_number": os.environ['JOB_ID'],
        "task_number": os.environ['SGE_TASK_ID'],
        "draw": parameters['input_draw'],
        "seed": parameters['random_seed'],
        "scenario": parameters['branch_configuration'],  # assumes leaves of branch config tree are scenarios
        "event": event,
        "exec_time": exec_time,
        "counters": (end - start).to_dict()
    }))
    logger.remove(perf_log)



import os
from pathlib import Path
from typing import Union


def mkdir(path: Union[str, Path], umask: int = 0o002, exists_ok: bool = False, parents: bool = False) -> None:
    """
    Utility method to create a directory with specified permissions
    :param path: path of the directory to create
    :param umask: umask specifying the desired permissions - defaults to 0o002
    :param exists_ok: if False, raises FileExistsError if the directory already exists
    :param parents: if False, raises FileNotFoundError if the directory's parent doesn't exist
    """
    path = Path(path)
    old_umask = os.umask(umask)
    try:
        path.mkdir(exist_ok=exists_ok, parents=parents)
    finally:
        os.umask(old_umask)



import json
import re
from pathlib import Path
from typing import Union

import numpy as np
import pandas as pd
import requests
from loguru import logger
from pandas.io.json import json_normalize


BASE_PERF_INDEX_COLS = ['host', 'job_number', 'task_number', 'draw', 'seed']

# The number of scenario columns beyond which we shorten the scenarios to a single string
COMPOUND_SCENARIO_COL_COUNT = 2


class PerformanceSummary:
    """
    A class to implement a getter for data in the workers' performance logs.

    Given a Path, a PerformanceSummary class provides a generator to get at each
    entry in the workers' performance logs. The class also provides a method
    to get all entries in a pd.DataFrame. This class is intended as a singleton
    to provide data about a single Vivarium simulation run.

    Attributes
    ----------
    log_dir : Path
        Path of log_dir

    Methods
    -------
    get_summaries():
        Generator to retrieve entries in the performance logs as dict objects.
    to_df():
        Returns the performance log data as a pd.DataFrame.
    """

    def __init__(self, log_dir: Path):
        self.log_dir: Path = log_dir
        self.errors: int = 0

    def get_summaries(self) -> dict:
        """Generator to get all performance summary log messages in PerformanceSummary"""
        for log in [f for f in self.log_dir.iterdir() if self.PERF_LOG_PATTERN.fullmatch(f.name)]:
            with log.open('r') as f:
                count: int = 0
                for line in f.readlines():
                    count += 1
                    try:
                        message = json.loads(line)['record']['message']
                    except Exception as e:
                        logger.warning(f"Exception: {e}. Malformed message in {log} line {count}, skipping...")
                        self.errors += 1
                        continue
                    m = self.TELEMETRY_PATTERN.fullmatch(str(message))
                    if m:
                        yield json_normalize(json.loads(message), sep='_')

    def to_df(self) -> pd.DataFrame:
        perf_data = []
        for item in self.get_summaries():
            perf_data.append(item)
        if len(perf_data) < 1:
            return pd.DataFrame()
        perf_df = pd.concat(perf_data)

        # Convert the Unix timestamps to datetimes
        for col in [col for col in perf_df.columns if col.startswith("event_")]:
            perf_df[col] = pd.to_datetime(perf_df[col], unit='s')

        # Remove trailing "_scenario" from normalized label
        perf_df.columns = perf_df.columns.str.replace('_scenario', '', regex=False)
        return perf_df

    TELEMETRY_PATTERN = re.compile(r'^{\"host\".+\"job_number\".+}$')
    PERF_LOG_PATTERN = re.compile(r'^perf\.([0-9]+)\.([0-9]+)\.log$')


def set_index_scenario_cols(perf_df: pd.DataFrame) -> (pd.DataFrame, list):
    """Given a dataframe from PerformanceSummary.to_df, add QPID Job API data for the job"""
    index_cols = BASE_PERF_INDEX_COLS
    scenario_cols = [col for col in perf_df.columns if col.startswith("scenario_")]
    index_cols.extend(scenario_cols)
    perf_df = perf_df.set_index(index_cols)
    return perf_df, scenario_cols


def add_jobapi_data(perf_df: pd.DataFrame):
    """Given a dataframe from PerformanceSummary.to_df, add QPID Job API data for the job.
    Job API reference: https://stash.ihme.washington.edu/projects/QPID/repos/job-db/browse/docs/index.md
    """
    try:
        job_numbers = perf_df['job_number'].unique()
        assert (len(job_numbers) == 1)
        jobapi_data = requests.get("http://jobapi.ihme.washington.edu/fair/queryjobids",
                                   params=[('job_number', job_numbers[0]), ('limit', 50000)]).json()
        jobapi_df = pd.DataFrame(jobapi_data["data"]).add_prefix('qpid_')
        perf_df = perf_df.astype({'job_number': np.int64, 'task_number': np.int64})
        perf_df = perf_df.merge(jobapi_df, left_on=['job_number', 'task_number'],
                                right_on=['qpid_job_number', 'qpid_task_number'])
    except Exception as e:
        logger.warning(f'Job API request failed with {e}')
    return perf_df


def print_stat_report(perf_df: pd.DataFrame, scenario_cols: list):
    """Print some helpful stats from the performance data, grouped by scenario_cols"""
    pd.set_option('display.max_rows', None)
    pd.set_option('display.max_columns', None)
    pd.options.display.float_format = '{:.2f}'.format

    do_compound = len(scenario_cols) > COMPOUND_SCENARIO_COL_COUNT

    perf_df = perf_df.reset_index()

    if do_compound:
        logger.info(f"compound scenario:\n({'/'.join([s.replace('scenario_', '') for s in scenario_cols])}):")
        perf_df["compound_scenario"] = perf_df[scenario_cols].to_csv(
            header=None, index=False, sep='/').strip('\n').split('\n')

    # Print execution times stats by scenario
    temp = (perf_df.set_index('compound_scenario' if do_compound else scenario_cols)
            .filter(like='exec_time_')
            .stack()
            .reset_index())

    if do_compound:
        cols = ['compound_scenario', 'measure', 'value']
    else:
        cols = scenario_cols
        cols.extend(['measure', 'value'])

    temp.columns = cols
    cols.remove('value')

    report_df = temp.groupby(cols).describe()
    report_df.columns = report_df.columns.droplevel()
    report_df = report_df.drop(['count', '25%', '50%', '75%'], axis=1)
    report_df = report_df.reset_index()

    # Abbreviate execution time measures for printing
    report_df['measure'] = report_df['measure'].replace('^exec_time_', '', regex=True)
    report_df['measure'] = report_df['measure'].replace('^simulant_initialization', 'sim_init', regex=True)
    report_df['measure'] = report_df['measure'].replace('minutes$', 'min', regex=True)
    report_df['measure'] = report_df['measure'].replace('seconds', 's', regex=True)

    report_df = report_df.set_index(cols).sort_index()
    logger.info(f'\n{report_df}')


def report_performance(input_directory: Union[Path, str], output_directory: Union[Path, str], output_hdf: bool,
                       verbose: int):
    """Main method for vipin reporting. Gets job performance data, outputs to a file, and logs a report."""
    input_directory, output_directory = Path(input_directory), Path(output_directory)
    perf_summary = PerformanceSummary(input_directory)

    perf_df = perf_summary.to_df()

    if len(perf_df) < 1:
        logger.warning(f'No performance data found in {input_directory}.')
        return  # nothing left to do

    # Add jobapi data about the job to dataframe
    perf_df = add_jobapi_data(perf_df)

    # Set index to include branch configuration/scenario columns
    perf_df, scenario_cols = set_index_scenario_cols(perf_df)

    # Write to file
    out_file = output_directory / 'log_summary'
    if output_hdf:
        out_file = out_file.with_suffix(".hdf")
        perf_df.to_hdf(out_file, key='worker_data')
    else:
        out_file = out_file.with_suffix(".csv")
        perf_df.to_csv(out_file)

    if verbose:
        print_stat_report(perf_df, scenario_cols)

    if perf_summary.errors > 0:
        logger.warning(f'{perf_summary.errors} log row{"s were" if perf_summary.errors > 1 else " was"} unreadable.')
    logger.info(f'Performance summary {"hdf" if output_hdf else "csv"} can be found at {out_file}, with '
                f'{perf_df.shape[0]} row{"s" if perf_df.shape[0] > 1 else ""}.')
    return



import os
import atexit
import math
import tempfile
import shutil
import socket
import subprocess
from typing import List, Dict, Tuple
from pathlib import Path
from time import sleep, time
import yaml

import numpy as np
import pandas as pd
from loguru import logger

from vivarium.framework.configuration import build_model_specification, ConfigurationError
from vivarium.framework.utilities import collapse_nested_dict
from vivarium.framework.artifact import parse_artifact_path_config

from vivarium_cluster_tools.psimulate.branches import Keyspace
from vivarium_cluster_tools.psimulate import globals as vct_globals, utilities
from vivarium_cluster_tools.psimulate.registry import RegistryManager
from vivarium_cluster_tools.vipin.perf_report import report_performance

drmaa = utilities.get_drmaa()

MODEL_SPEC_FILENAME = 'model_specification.yaml'
FULL_ARTIFACT_PATH_KEY = f'{vct_globals.INPUT_DATA_KEY}.{vct_globals.ARTIFACT_PATH_KEY}'


class RunContext:
    def __init__(self, model_specification_file: str, branch_configuration_file: str, artifact_path: str,
                 output_directory: Path, logging_directories: Dict[str, Path], num_input_draws: int,
                 num_random_seeds: int, restart: bool, expand: Dict[str, int], no_batch: bool):
        self.number_already_completed = 0
        self.output_directory = output_directory
        self.no_batch = no_batch
        self.sge_log_directory = logging_directories['sge']
        self.worker_log_directory = logging_directories['worker']

        if restart:
            self.keyspace = Keyspace.from_previous_run(self.output_directory)
            self.existing_outputs = pd.read_hdf(self.output_directory / 'output.hdf')
            if expand:
                self.keyspace.add_draws(expand['num_draws'])
                self.keyspace.add_seeds(expand['num_seeds'])
                self.keyspace.persist(self.output_directory)
        else:
            model_specification = build_model_specification(model_specification_file)

            self.keyspace = Keyspace.from_branch_configuration(num_input_draws, num_random_seeds,
                                                               branch_configuration_file)
            if artifact_path:
                if FULL_ARTIFACT_PATH_KEY in self.keyspace:
                    raise ConfigurationError('Artifact path cannot be specified both in the branch specification file'
                                             ' and as a command line argument.', artifact_path)
                if not Path(artifact_path).exists():
                    raise FileNotFoundError(f"Cannot find artifact at path {artifact_path}")

            elif vct_globals.ARTIFACT_PATH_KEY in model_specification.configuration[vct_globals.INPUT_DATA_KEY]:
                artifact_path = parse_artifact_path_config(model_specification.configuration)

            if artifact_path:
                model_specification.configuration[vct_globals.INPUT_DATA_KEY].update(
                    {vct_globals.ARTIFACT_PATH_KEY: artifact_path},
                    source=__file__)

            with open(self.output_directory / MODEL_SPEC_FILENAME, 'w') as config_file:
                yaml.dump(model_specification.to_dict(), config_file)

            self.existing_outputs = None

            # Log some basic stuff about the simulation to be run.
            self.keyspace.persist(self.output_directory)
        self.model_specification = self.output_directory / MODEL_SPEC_FILENAME


class NativeSpecification:
    def __init__(self, project: str, queue: str, peak_memory: int, max_runtime: str, job_name: str, **__):

        try:
            self.cluster_name = os.environ['SGE_CLUSTER_NAME']
        except KeyError:
            raise RuntimeError('This tool must be run from an SGE/UGE cluster.')

        self.project = project
        self.peak_memory = peak_memory
        self.max_runtime = max_runtime
        self.job_name = job_name
        self.queue = self.get_queue(queue, max_runtime)
        self.threads = vct_globals.DEFAULT_THREADS_PER_JOB
        self.qsub_validation = 'n'

        self.flag_map = {
            'project': '-P {value}',
            'peak_memory': '-l m_mem_free={value}G',
            'max_runtime': '-l h_rt={value}',
            'threads': '-l fthread={value}',
            'job_name': '-N {value}',
            'queue': '-q {value}',
            'qsub_validation': '-w {value}'
        }

        # base configuration resources
        self.allowed_resources = ['job_name', 'queue', 'max_runtime', 'qsub_validation']

        # IHME-specific configuration
        if self.cluster_name == 'cluster':
            self.allowed_resources.extend(['project', 'peak_memory', 'threads'])

    def get_queue(self, queue: str, max_runtime: str) -> str:
        valid_queues = self.get_valid_queues(max_runtime)
        if queue is None:
            return valid_queues[0]
        else:
            if queue in valid_queues:
                return queue
            else:
                raise ValueError("Specified queue is not valid for jobs with the max runtime provided. "
                                 "Likely you requested all.q for a job requiring long.q.")

    @staticmethod
    def get_valid_queues(max_runtime: str) -> List[str]:
        runtime_args = max_runtime.split(":")

        if len(runtime_args) != 3:
            raise ValueError("Invalid --max-runtime supplied. Format should be hh:mm:ss.")
        else:
            hours, minutes, seconds = runtime_args

        runtime_in_hours = int(hours) + float(minutes) / 60. + float(seconds) / 3600.

        if runtime_in_hours <= vct_globals.ALL_Q_MAX_RUNTIME_HOURS:
            return ['all.q', 'long.q']
        elif runtime_in_hours <= vct_globals.LONG_Q_MAX_RUNTIME_HOURS:
            return ['long.q']
        else:
            raise ValueError(f"Max runtime value too large. Must be less than {vct_globals.LONG_Q_MAX_RUNTIME_HOURS}h.")

    def __str__(self):
        return " ".join([self.flag_map[resource].format(value=getattr(self, resource))
                         for resource in self.allowed_resources])


def init_job_template(jt, native_specification: NativeSpecification, sge_log_directory: Path,
                      worker_log_directory: Path, worker_settings_file: Path):
    launcher = tempfile.NamedTemporaryFile(mode='w', dir='.', prefix='vivarium_cluster_tools_launcher_',
                                           suffix='.sh', delete=False)
    atexit.register(lambda: os.remove(launcher.name))
    output_dir = str(worker_settings_file.resolve().parent)
    launcher.write(f'''
    export VIVARIUM_LOGGING_DIRECTORY={worker_log_directory}
    export PYTHONPATH={output_dir}:$PYTHONPATH

    {shutil.which('rq')} worker -c {worker_settings_file.stem} --name ${{JOB_ID}}.${{SGE_TASK_ID}} --burst \
        -w "vivarium_cluster_tools.psimulate.distributed_worker.ResilientWorker" \
        --exception-handler "vivarium_cluster_tools.psimulate.distributed_worker.retry_handler" vivarium

    ''')
    launcher.close()

    jt.workingDirectory = os.getcwd()
    jt.remoteCommand = shutil.which('sh')
    jt.args = [launcher.name]
    jt.outputPath = f":{sge_log_directory}"
    jt.errorPath = f":{sge_log_directory}"
    jt.jobEnvironment = {
        'LC_ALL': 'en_US.UTF-8',
        'LANG': 'en_US.UTF-8',
    }
    jt.joinFiles = True
    jt.nativeSpecification = str(native_specification)
    return jt


def get_random_free_port() -> int:
    # NOTE: this implementation is vulnerable to rare race conditions where some other process gets the same
    # port after we free our socket but before we use the port number we got. Should be so rare in practice
    # that it doesn't matter.
    s = socket.socket()
    s.bind(('', 0))
    port = s.getsockname()[1]
    s.close()
    return port


def launch_redis(port: int) -> subprocess.Popen:
    try:
        # inline config for redis server.
        redis_process = subprocess.Popen(["redis-server", "--port", f"{port}",
                                          "--timeout", "2",
                                          "--protected-mode", "no"], stdout=subprocess.DEVNULL,
                                         stderr=subprocess.DEVNULL)
    except FileNotFoundError:
        raise OSError("In order for redis to launch you need both the redis client and the python bindings. "
                      "You seem to be missing the redis client.  Do 'conda install redis' and try again. If "
                      "failures continue you may need to download redis yourself, make it and add it to PATH.")
    atexit.register(redis_process.kill)
    return redis_process


def launch_redis_processes(num_processes: int) -> (str, List[Tuple[str, int]]):
    hostname = socket.getfqdn()
    redis_ports = []
    for i in range(num_processes):
        port = get_random_free_port()
        logger.info(f'Starting Redis Broker at {hostname}:{port}')
        launch_redis(port)
        redis_ports.append((hostname, port))

    redis_urls = [f'redis://{hostname}:{port}' for hostname, port in redis_ports]
    worker_config = f"import random\nredis_urls = {redis_urls}\nREDIS_URL = random.choice(redis_urls)\n\n"
    return worker_config, redis_ports


def start_cluster(drmaa_session, num_workers: int, sge_log_directory: Path, worker_log_directory: Path,
                  worker_settings_file: Path, native_specification: NativeSpecification):
    s = drmaa_session
    jt = init_job_template(s.createJobTemplate(), native_specification, sge_log_directory,
                           worker_log_directory, worker_settings_file)
    if num_workers:
        job_ids = s.runBulkJobs(jt, 1, num_workers, 1)
        array_job_id = job_ids[0].split('.')[0]

        def kill_jobs():
            if "drmaa" not in dir():
                # FIXME: The global drmaa should be available here.
                #        This is maybe a holdover from old code?
                #        Maybe something to do with atexit?
                drmaa = utilities.get_drmaa()

            try:
                s.control(array_job_id, drmaa.JobControlAction.TERMINATE)
            # FIXME: Hack around issue where drmaa.errors sometimes doesn't
            #        exist.
            except Exception as e:
                if 'There are no jobs registered' in str(e):
                    # This is the case where all our workers have already shut down
                    # on their own, which isn't actually an error.
                    pass
                elif 'Discontinued delete' in str(e):
                    # sge has already cleaned up some of the jobs.
                    pass
                else:
                    raise

        atexit.register(kill_jobs)


def build_job_list(ctx: RunContext) -> List[dict]:
    jobs = []
    number_already_completed = 0

    for (input_draw, random_seed, branch_config) in ctx.keyspace:
        parameters = {'model_specification_file': str(ctx.model_specification),
                      'branch_configuration': branch_config,
                      'input_draw': int(input_draw),
                      'random_seed': int(random_seed),
                      'results_path': ctx.output_directory,
                      }

        do_schedule = True
        if ctx.existing_outputs is not None:
            mask = ctx.existing_outputs.input_draw == int(input_draw)
            mask &= ctx.existing_outputs.random_seed == int(random_seed)
            if branch_config:
                for k, v in collapse_nested_dict(branch_config):
                    if isinstance(v, float):
                        mask &= np.isclose(ctx.existing_outputs[k], v)
                    else:
                        mask &= ctx.existing_outputs[k] == v
            do_schedule = not np.any(mask)

        if do_schedule:
            jobs.append(parameters)
        else:
            number_already_completed += 1

    if number_already_completed:
        logger.info(f"{number_already_completed} of {len(ctx.keyspace)} jobs completed in previous run.")
        if number_already_completed != len(ctx.existing_outputs):
            logger.warning("There are jobs from the previous run which would not have been created "
                           "with the configuration saved with the run. That either means that code "
                           "has changed between then and now or that the outputs or configuration data "
                           "have been modified. This may represent a serious error so give it some thought.")

    ctx.number_already_completed = number_already_completed
    np.random.shuffle(jobs)
    return jobs


def concat_preserve_types(df_list: List[pd.DataFrame]) -> pd.DataFrame:
    """Concatenation preserves all ``numpy`` dtypes but does not preserve any
    pandas specific dtypes (e.g., categories become objects."""
    dtypes = df_list[0].dtypes
    columns_by_dtype = [list(dtype_group.index) for _, dtype_group in dtypes.groupby(dtypes)]

    splits = []
    for columns in columns_by_dtype:
        slices = [df.filter(columns) for df in df_list]
        splits.append(pd.DataFrame(data=np.concatenate(slices), columns=columns))
    return pd.concat(splits, axis=1)


def concat_results(old_results: pd.DataFrame, new_results: List[pd.DataFrame]) -> pd.DataFrame:
    # Skips all the pandas index checking because columns are in the same order.
    start = time()

    to_concat = [d.reset_index(drop=True) for d in new_results]
    if not old_results.empty:
        to_concat += [old_results.reset_index(drop=True)]

    results = concat_preserve_types(to_concat)

    end = time()
    logger.info(f"Concatenated {len(new_results)} results in {end - start:.2f}s.")
    return results


def write_results_batch(ctx: RunContext, written_results: pd.DataFrame, unwritten_results: List[pd.DataFrame],
                        batch_size: int = 50) -> (pd.DataFrame, List[pd.DataFrame]):
    new_results_to_write, unwritten_results = (unwritten_results[:batch_size], unwritten_results[batch_size:])
    results_to_write = concat_results(written_results, new_results_to_write)

    start = time()
    retries = 3
    while retries:
        try:
            output_path = ctx.output_directory / 'output.hdf'
            # Writing to an hdf over and over balloons the file size so write to new file and move it over to avoid
            temp_output_path = output_path.with_name(output_path.name + 'update')
            results_to_write.to_hdf(temp_output_path, 'data')
            temp_output_path.replace(output_path)
            break
        except Exception:
            logger.warning(f'Error trying to write results to hdf, retries remaining {retries}')
            sleep(30)
            retries -= 1
    end = time()
    logger.info(f"Updated output.hdf in {end - start:.4f}s.")
    return results_to_write, unwritten_results


def process_job_results(registry_manager: RegistryManager, ctx: RunContext):
    start_time = time()

    if ctx.existing_outputs is not None:
        written_results = ctx.existing_outputs
    else:
        written_results = pd.DataFrame()
    unwritten_results = []

    logger.info('Entering main processing loop.')
    batch_size = 200
    try:
        while registry_manager.jobs_to_finish:
            sleep(5)
            unwritten_results.extend(registry_manager.get_results())
            if ctx.no_batch and unwritten_results:
                written_results, unwritten_results = write_results_batch(ctx, written_results,
                                                                        unwritten_results, len(unwritten_results))
            elif len(unwritten_results) > batch_size:
                written_results, unwritten_results = write_results_batch(ctx, written_results,
                                                                        unwritten_results, batch_size)

            registry_manager.update_and_report()
            logger.info(f'Unwritten results: {len(unwritten_results)}')
            logger.info(f'Elapsed time: {(time() - start_time)/60:.1f} minutes.')
    finally:
        batch_size = 500
        while unwritten_results:
            written_results, unwritten_results = write_results_batch(ctx, written_results, unwritten_results,
                                                                    batch_size=batch_size)
            logger.info(f'Unwritten results: {len(unwritten_results)}')
            logger.info(f'Elapsed time: {(time() - start_time) / 60:.1f} minutes.')


def check_user_sge_config():
    """Warn if a user has set their stdout and stderr output paths
    in a home directory config file. This overrides settings from py-drmaa."""

    sge_config = Path().home() / ".sge_request"

    if sge_config.exists():
        with sge_config.open('r') as f:
            for line in f:
                line = line.strip()
                if (('-o ' in line) or ('-e' in line)) and not line.startswith("#"):
                    logger.warning("You may have settings in your .sge_request file "
                                   "that could overwrite the log location set by this script. "
                                   f"Your .sge_request file is here: {sge_config}.  Look for "
                                   "-o and -e and comment those lines to receive logs side-by-side"
                                   "with the worker logs.")


def try_run_vipin(log_path: Path):

    try:
        report_performance(input_directory=log_path, output_directory=log_path, output_hdf=False, verbose=1)
    except Exception as e:
        logger.warning(f'Performance reporting failed with: {e}')


def main(model_specification_file: str, branch_configuration_file: str, artifact_path: str, result_directory: str,
         native_specification: dict, redis_processes: int, num_input_draws: int = None,
         num_random_seeds: int = None, restart:  bool = False,
         expand: Dict[str, int] = None, no_batch: bool = False, no_cleanup: bool = False):

    utilities.exit_if_on_submit_host(utilities.get_hostname())

    output_dir, logging_dirs = utilities.setup_directories(model_specification_file, result_directory,
                                                           restart, expand=bool(num_input_draws or num_random_seeds))

    if not no_cleanup:
        atexit.register(utilities.check_for_empty_results_dir, output_dir=output_dir)

    atexit.register(lambda: logger.remove())

    native_specification['job_name'] = output_dir.parts[-2]
    native_specification = NativeSpecification(**native_specification)

    utilities.configure_master_process_logging_to_file(logging_dirs['main'])
    utilities.validate_environment(output_dir)

    ctx = RunContext(model_specification_file, branch_configuration_file, artifact_path, output_dir, logging_dirs,
                     num_input_draws, num_random_seeds, restart, expand, no_batch)
    check_user_sge_config()
    jobs = build_job_list(ctx)

    if len(jobs) == 0:
        logger.info("Nothing to do")
        return

    logger.info('Starting jobs. Results will be written to: {}'.format(ctx.output_directory))

    if redis_processes == -1:
        redis_processes = int(math.ceil(len(jobs) / vct_globals.DEFAULT_JOBS_PER_REDIS_INSTANCE))

    worker_template, redis_ports = launch_redis_processes(redis_processes)
    worker_file = output_dir / 'settings.py'
    with worker_file.open('w') as f:
        f.write(worker_template)

    registry_manager = RegistryManager(redis_ports, ctx.number_already_completed)
    registry_manager.enqueue(jobs)

    drmaa_session = drmaa.Session()
    drmaa_session.initialize()

    start_cluster(drmaa_session,
                  len(jobs),
                  ctx.sge_log_directory,
                  ctx.worker_log_directory,
                  worker_file,
                  native_specification)

    process_job_results(registry_manager, ctx)

    try_run_vipin(logging_dirs['worker'])

    logger.info('Jobs completed. Results written to: {}'.format(ctx.output_directory))



import os
import sys
from datetime import datetime
from pathlib import Path
from shutil import rmtree
from typing import Union


from loguru import logger
from typing import Dict, List, Sequence

# depending on version of pip, freeze may be in one of two places
try:
    from pip._internal.operations import freeze
except ImportError:
    from pip.operations import freeze

from vivarium_cluster_tools import utilities as vct_utils
from vivarium_cluster_tools.psimulate import globals as vct_globals
from vivarium_cluster_tools.vipin.perf_report import report_performance


def get_drmaa() -> object:
    try:
        import drmaa
    except (RuntimeError, OSError):
        if 'SGE_CLUSTER_NAME' in os.environ:
            os.environ['DRMAA_LIBRARY_PATH'] = '/opt/sge/lib/lx-amd64/libdrmaa.so'
            import drmaa
        else:
            drmaa = object()
    return drmaa


def add_logging_sink(sink, verbose: int, colorize=False, serialize=False):
    message_format = ('<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | '
                      '<cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> '
                      '- <level>{message}</level>')
    if verbose == 0:
        def quiet_filter(record):
            return record.get('extra', {}).get('queue', None) == 'all'

        logger.add(sink, colorize=colorize, level="INFO", format=message_format,
                   filter=quiet_filter, serialize=serialize)
    elif verbose == 1:
        logger.add(sink, colorize=colorize, level="INFO", format=message_format, serialize=serialize)
    elif verbose >= 2:
        logger.add(sink, colorize=colorize, level="DEBUG", format=message_format, serialize=serialize)


def configure_master_process_logging_to_terminal(verbose: int):
    logger.remove(0)  # Clear default configuration
    add_logging_sink(sys.stdout, verbose, colorize=True)


def configure_master_process_logging_to_file(output_directory: Path):
    master_log = output_directory / 'master.log'
    serial_log = output_directory / 'master.log.json'
    add_logging_sink(master_log, verbose=2)
    add_logging_sink(serial_log, verbose=2, serialize=True)


def get_output_directory(model_specification_file: str = None,
                         output_directory: str = None, restart: bool = False) -> Path:
    if restart:
        output_directory = Path(output_directory)
    else:
        root = Path(output_directory) if output_directory else Path(vct_globals.DEFAULT_OUTPUT_DIRECTORY)
        launch_time = datetime.now().strftime("%Y_%m_%d_%H_%M_%S")
        model_specification_name = Path(model_specification_file).stem
        output_directory = root / model_specification_name / launch_time
    return output_directory


def set_permissions(output_dir: Path):
    """Call to achieve side effect of relaxing permissions to 775
        on output dir and the parent"""
    permissions = 0o775
    output_dir.parent.chmod(permissions)
    output_dir.chmod(permissions)


def setup_directories(model_specification_file: str, result_directory: str,
                      restart: bool, expand: bool) -> (Path, Dict[str, Path]):
    output_directory = get_output_directory(model_specification_file, result_directory, restart)

    if restart and not expand:
        command = 'restart'
        launch_time = datetime.now().strftime("%Y_%m_%d_%H_%M_%S")
    elif restart and expand:
        command = 'expand'
        launch_time = datetime.now().strftime("%Y_%m_%d_%H_%M_%S")
    else:
        command = 'run'
        launch_time = output_directory.stem

    logging_directory = output_directory / "logs" / f'{launch_time}_{command}'

    logging_dirs = {'main': logging_directory,
                    'sge': logging_directory / 'sge_logs',
                    'worker': logging_directory / 'worker_logs'}

    vct_utils.mkdir(output_directory, exists_ok=True, parents=True)
    for d in logging_dirs.values():
        vct_utils.mkdir(d, parents=True)

    return output_directory, logging_dirs


def get_hostname() -> str:
    return os.environ.get(vct_globals.CLUSTER_ENV_HOSTNAME)


def exit_if_on_submit_host(name: str):
    if vct_globals.SUBMIT_HOST_MARKER in name:
        raise RuntimeError('This tool must not be run from a submit host.')


def chunks(l: Sequence, n: int):
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(l), n):
        yield l[i:i + n]


def parse_package_version(s: str) -> (str, str):
    if 'no version control' in s:  # installed from non-git controlled source code in editable mode
        s = s.split('(')[1].split(')')[0]  # yields <package>==<version>

    packaged = s.split('==')
    if len(packaged) == 2:  # not installed from source code, e.g., <package>==<version>
        package = packaged[0]
        version = packaged[1]
    else:  # installed from source code in editable mode so more parsing required
        sourced = s.split('=')
        package = sourced[-1]
        version = sourced[0].split('@')[-1].replace('#egg', '')  # version here is git hash
    return package, version


def convert_pip_list_to_dict(pf_list: List[str]) -> Dict:
    return {p: v for p, v in [parse_package_version(s) for s in pf_list]}


def compare_environments(current: Dict, original: Dict):
    differences = []

    current_packages = set(current.keys())
    original_packages = set(original.keys())

    new_packages = current_packages.difference(original_packages)
    if new_packages:
        differences.append(
            f'The current environment contains the following packages not present in the original '
            f'environment: {new_packages}.')

    missing_packages = original_packages.difference(current_packages)
    if missing_packages:
        differences.append(
            f'The current environment is missing the following packages present in the original '
            f'environment: {missing_packages}.')

    differing_versions = []
    for p in current_packages.intersection(original_packages):
        if current[p] != original[p]:
            differing_versions.append(f'{p}: {original[p]} -> {current[p]}')

    if differing_versions:
        differences.append(
            f'Different versions found in current and original environment for the following '
            f'packages: {differing_versions}.')

    if differences:
        differences = "\n".join(differences)
        raise ValueError(f'Differences found between environment used for original run and current environment. '
                         f'In order to successfully run, you should make a new environment using the requirements.txt '
                         f'file found in the output directory. Differences found as follows: {differences}.')


def validate_environment(output_dir: Path):
    original_environment_file = output_dir / 'requirements.txt'

    current_environment_list = [p for p in freeze.freeze()]
    if not original_environment_file.exists():  # original run
        with original_environment_file.open('w') as f:
            f.write('\n'.join(current_environment_list))
    else:  # compare with original
        with original_environment_file.open('r') as f:
            original_environment_list = [line.replace('\n', '') for line in f]
        current_environment = convert_pip_list_to_dict(current_environment_list)
        original_environment = convert_pip_list_to_dict(original_environment_list)
        compare_environments(current_environment, original_environment)
        logger.info('Validation of environment successful. All pip installed packages match '
                    'original versions. Run can proceed.')


def check_for_empty_results_dir(output_dir: Path):
    """Remove the results directory including runner and worker logs if the simulation produced no results (i.e.,
    it failed). """
    if not (output_dir / 'output.hdf').exists():
        rmtree(output_dir)



import click

from vivarium_cluster_tools.vipin import utilities, perf_report


@click.command()
@click.argument('logs-directory', type=click.Path(exists=True, file_okay=False))
@click.option('--result-directory', '-o', type=click.Path(exists=True, file_okay=False),
              help='The directory into which to write the summary of the parsed logs. '
                   'Defaults to given logs directory if not given.')
@click.option('--hdf/--csv', default=False,
              help='Choose hdf or csv for output data. Defaults to csv.')
@click.option('-v', 'verbose', count=True, help='Configure logging verbosity.')
def vipin(logs_directory, result_directory, hdf, verbose):
    """Get performance information from worker_logs from a ``psimulate`` command.

    Given a worker logs directory from a previous run, a summary csv will be
    created in the ``result_directory`` (which defaults to the given logs
    directory unless otherwise specified).
    """
    utilities.configure_master_process_logging_to_terminal(verbose)
    if not result_directory:
        result_directory = logs_directory

    perf_report.report_performance(logs_directory, result_directory, hdf, verbose)




from itertools import product

import yaml
import numpy as np
from pathlib import Path

from vivarium.framework.utilities import collapse_nested_dict

from vivarium_cluster_tools.psimulate import globals as vct_globals

FULL_ARTIFACT_PATH_KEY = f'{vct_globals.INPUT_DATA_KEY}.{vct_globals.ARTIFACT_PATH_KEY}'


class Keyspace:
    """A representation of a collection of simulation configurations."""

    def __init__(self, branches, keyspace):
        self.branches = branches
        self._keyspace = keyspace

    @classmethod
    def from_branch_configuration(cls, num_input_draws, num_random_seeds, branch_configuration_file):
        """
        Parameters
        ----------
        num_input_draws: int
            The number of GBD input draws to run.
        num_random_seeds: int
            The number of different random seeds to use for each GBD input draw.  Each model
            draw creates a simulation with fixed GBD data, but a different sample of the
            exogenous randomness used in the simulation.
        branch_configuration_file: str
            Absolute path to the branch configuration file.
        """
        branches, input_draw_count, random_seed_count = load_branches(num_input_draws,
                                                                      num_random_seeds,
                                                                      branch_configuration_file)
        keyspace = calculate_keyspace(branches)
        keyspace['input_draw'] = calculate_input_draws(input_draw_count)
        keyspace['random_seed'] = calculate_random_seeds(random_seed_count)

        return Keyspace(branches, keyspace)

    @classmethod
    def from_previous_run(cls, path: Path):
        keyspace = yaml.full_load((path / 'keyspace.yaml').read_text())
        branches = yaml.full_load((path / 'branches.yaml').read_text())
        return Keyspace(branches, keyspace)

    def get_data(self):
        """Returns a copy of the underlying keyspace data."""
        return self._keyspace.copy()

    def get_branch_number(self, branch):
        for (i, b) in enumerate(self.branches):
            if b == branch:
                return i
        raise KeyError(f"No matching branch {branch}")

    def persist(self, output_directory: Path):
        (output_directory / 'keyspace.yaml').write_text(yaml.dump(self.get_data()))
        (output_directory / 'branches.yaml').write_text(yaml.dump(self.branches))

    def add_draws(self, num_draws):
        existing = self._keyspace['input_draw']
        additional = calculate_input_draws(num_draws, existing)
        self._keyspace['input_draw'] = existing + additional

    def add_seeds(self, num_seeds):
        existing = self._keyspace['random_seed']
        additional = calculate_random_seeds(num_seeds, existing)
        self._keyspace['random_seed'] = existing + additional

    def __contains__(self, item):
        """Checks whether the item is present in the Keyspace"""
        return item in self._keyspace

    def __iter__(self):
        """Yields and individual simulation configuration from the keyspace."""
        for job_config in product(self._keyspace['input_draw'], self._keyspace['random_seed'], self.branches):
            yield job_config

    def __len__(self):
        """Returns the number of individual simulation runs this keyspace represents."""
        return len(list(product(self._keyspace['input_draw'], self._keyspace['random_seed'], self.branches)))


def calculate_input_draws(input_draw_count, existing_draws=None):
    """Determines a random sample of the GBD input draws to use given a draw
    count and any existing draws.

    Parameters
    ----------
    input_draw_count: int
        The number of draws to pull.

    existing_draws: list
        Any draws that have already been pulled and should not be pulled again.

    Returns
    -------
    Iterable:
        A set of unique input draw numbers, guaranteed not to overlap with any
        existing draw numbers.
    """
    np.random.seed(123456)
    if existing_draws:
        possible = list(set(range(1000)).difference(existing_draws))
        min_input_draw_count_allowed = 0
    else:
        possible = list(range(1000))
        min_input_draw_count_allowed = 1

    if min_input_draw_count_allowed <= input_draw_count <= len(possible):
        return np.random.choice(possible, input_draw_count, replace=False).tolist()
    else:
        raise ValueError(f"Input draw count must be between {min_input_draw_count_allowed} "
                         f"and {len(possible)} (inclusive). You specified {input_draw_count}.")


def calculate_random_seeds(random_seed_count, existing_seeds=None):
    """Generates random seeds to use given a count of seeds and any existing
    seeds.

    Parameters
    ----------
    random_seed_count: int
        The number of random seeds to generate.

    existing_seeds: list
        Any random seeds that have already been generated and should not be
        generated again.

    Returns
    -------
    Iterable:
        A set of unique random seeds, guaranteed not to overlap with any
        existing random seeds.
    """
    if not random_seed_count:
        return []

    np.random.seed(654321)

    if existing_seeds:
        min_possible = max(existing_seeds) + 1
    else:
        existing_seeds = []
        min_possible = 0

    low, high = min_possible, min_possible + 10*random_seed_count
    possible = list(range(low, high))
    return np.random.choice(possible, random_seed_count, replace=False).tolist()


def calculate_keyspace(branches):
    if branches[0] is None:
        return {}

    keyspace = {k: {v} for k, v in collapse_nested_dict(branches[0])}

    for branch in branches[1:]:
        branch = dict(collapse_nested_dict(branch))
        if set(branch.keys()) != set(keyspace.keys()):
            raise ValueError("All branches must have the same keys")
        for k, v in branch.items():
            if k == FULL_ARTIFACT_PATH_KEY:
                validate_artifact_path(v)
            keyspace[k].add(v)
    keyspace = {k: list(v) for k, v in keyspace.items()}
    return keyspace


def load_branches(num_input_draws, num_random_seeds, branch_configuration_file):
    if num_input_draws is None and num_random_seeds is None and branch_configuration_file is not None:
        input_draw_count, random_seed_count, branches = load_branch_configurations(branch_configuration_file)
    elif num_input_draws is not None and branch_configuration_file is None:
        input_draw_count = num_input_draws
        random_seed_count = num_random_seeds if num_random_seeds is not None else 1
        branches = [None]
    else:
        raise ValueError('Must supply one of branch_configuration_file or --num_input_draws but not both')

    return branches, input_draw_count, random_seed_count


def load_branch_configurations(path):
    with open(path) as f:
        data = yaml.full_load(f)

    input_draw_count = data.get('input_draw_count', 1)
    random_seed_count = data.get('random_seed_count', 1)

    assert input_draw_count <= 1000, "Cannot use more that 1000 draws from GBD"

    if 'branches' in data:
        branches = expand_branch_templates(data['branches'])
    else:
        branches = [None]

    return input_draw_count, random_seed_count, branches


def expand_branch_templates(templates):
    """
    Take a list of dictionaries of configuration values (like the ones used in
    experiment branch configurations) and expand it by taking any values which
    are lists and creating a new set of branches which is made up of the
    product of all those lists plus all non-list values.

    For example this:

    {'a': {'b': [1,2], 'c': 3, 'd': [4,5,6]}}

    becomes this:

    [
        {'a': {'b': 1, 'c': 3, 'd': 4}},
        {'a': {'b': 2, 'c': 3, 'd': 5}},
        {'a': {'b': 1, 'c': 3, 'd': 6}},
        {'a': {'b': 2, 'c': 3, 'd': 4}},
        {'a': {'b': 1, 'c': 3, 'd': 5}},
        {'a': {'b': 2, 'c': 3, 'd': 6}}
    ]

    """
    expanded_branches = []

    for branch in templates:
        branch = sorted(collapse_nested_dict(branch))
        branch = [(k, v if isinstance(v, list) else [v]) for k, v in branch]
        expanded_size = np.product([len(v) for k, v in branch])
        new_branches = []
        pointers = {k: 0 for k, _ in branch}
        for _ in range(expanded_size):
            new_branch = []
            tick = True
            for k, v in branch:
                new_branch.append((k, v[pointers[k]]))
                if tick:
                    i = pointers[k]+1
                    if i < len(v):
                        tick = False
                        pointers[k] = i
                    else:
                        pointers[k] = 0
            new_branches.append(new_branch)
        expanded_branches.extend(new_branches)

    final_branches = []
    for branch in expanded_branches:
        root = {}
        final_branches.append(root)
        for k, v in branch:
            current = root
            *ks, k = k.split('.')
            for sub_k in ks:
                if sub_k in current:
                    current = current[sub_k]
                else:
                    current[sub_k] = {}
                    current = current[sub_k]
            current[k] = v

    return final_branches


def validate_artifact_path(artifact_path: str) -> str:
    """Validates that the path to the data artifact from the branches file exists.

    The path specified in the configuration must be absolute

    Parameters
    ----------
    artifact_path :
        The path to the artifact.
    """
    path = Path(artifact_path)

    if not path.is_absolute() or not path.exists():
        raise FileNotFoundError(f"Cannot find artifact at path {path}")

    return str(path)





# Cluster projects under the purview of Simulation Science
CLUSTER_PROJECTS = ['proj_cost_effect', 'proj_cost_effect_diarrhea', 'proj_cost_effect_dcpn', 'proj_cost_effect_conic',
                    'proj_csu']
DEFAULT_CLUSTER_PROJECT = 'proj_cost_effect'

DEFAULT_OUTPUT_DIRECTORY = '/share/costeffectiveness/results'

DEFAULT_JOBS_PER_REDIS_INSTANCE = 1000
DEFAULT_THREADS_PER_JOB = 1

# Cluster specific parameters
ALL_Q_MAX_RUNTIME_HOURS = 3 * 24
LONG_Q_MAX_RUNTIME_HOURS = 16 * 24

CLUSTER_ENV_HOSTNAME = 'HOSTNAME'
SUBMIT_HOST_MARKER = '-submit-'


# Configuration key constants
INPUT_DATA_KEY = 'input_data'
ARTIFACT_PATH_KEY = 'artifact_path'



import os
from setuptools import setup, find_packages

if __name__ == "__main__":

    base_dir = os.path.dirname(__file__)
    src_dir = os.path.join(base_dir, "src")

    about = {}
    with open(os.path.join(src_dir, "vivarium_cluster_tools", "__about__.py")) as f:
        exec(f.read(), about)

    with open(os.path.join(base_dir, "README.rst")) as f:
        long_description = f.read()

    install_requires = [
        'pandas',
        'numpy',
        'tables',
        'loguru',
        'pyyaml>=5.1',
        'drmaa',
        'redis<=3.0.1',
        'rq>=1.0, <=1.2.2',
        'vivarium>=0.9.3',
        'click',
        'psutil',
        'requests'
    ]

    test_requirements = [
        'pytest',
        'pytest-mock',
    ]

    doc_requirements = [
        'sphinx<2.1',
        'sphinx-autodoc-typehints<=1.7',
        'sphinx-rtd-theme',
        'sphinx-click',
        'IPython',
        'matplotlib'
    ]

    setup(
        name=about['__title__'],
        version=about['__version__'],

        description=about['__summary__'],
        long_description=long_description,
        url=about["__uri__"],

        author=about["__author__"],
        author_email=about["__email__"],

        package_dir={'': 'src'},
        packages=find_packages(where='src'),
        include_package_data=True,
        entry_points="""
            [console_scripts]
            psimulate=vivarium_cluster_tools.psimulate.cli:psimulate
            vipin=vivarium_cluster_tools.vipin.cli:vipin
        """,

        install_requires=install_requires,
        extras_require={
            'docs': doc_requirements,
            'test': test_requirements,
            'dev': doc_requirements + test_requirements,
        },

        zip_safe=False,
    )



__all__ = [
    "__title__", "__summary__", "__uri__", "__version__", "__author__",  "__email__",
]

__title__ = "vivarium_cluster_tools"
__summary__ = "A set of tools for running simulation using vivarium on cluster."
__uri__ = "https://github.com/ihmeuw/vivarium_cluster_tools"

__version__ = "1.2.4"

__author__ = "The vivarium developers"
__email__ = "vivarium.dev@gmail.com"



import sys

from loguru import logger


def add_logging_sink(sink, verbose, colorize=False, serialize=False):
    message_format = ('<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | '
                      '<cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> '
                      '- <level>{message}</level>')
    if verbose == 0:
        logger.add(sink, colorize=colorize, level="INFO", format=message_format, serialize=serialize)
    elif verbose >= 1:
        logger.add(sink, colorize=colorize, level="DEBUG", format=message_format, serialize=serialize)


def configure_master_process_logging_to_terminal(verbose):
    logger.remove()  # Clear default configuration
    add_logging_sink(sys.stdout, verbose, colorize=True)
